import numpy as np


def create_univariate_windows(df, train_horizon, pred_horizon, column=None):
    """Given a dataframe or raw data sequence, this will create deep learning friendly
    matrices (X, y) with a given training and prediction length.

    Args:
        df (array_like): raw data potentially with column names if a full dataframe
        train_horizon (int): number of data points in each training observation of the X matrix
        pred_horizon (int): number of data points in each prediction observation of the y matrix
        column (str, optional): column to use if df is a pandas dataframe. Defaults to None.
    """
    # number of windows
    num_windows = df.shape[0] - train_horizon - pred_horizon + 1

    # allocate memory to numpy
    X = np.zeros([num_windows, train_horizon, 1])
    y = np.zeros([num_windows, pred_horizon])

    # create numpy object
    if column is not None:
        raw_data = df[[column]].to_numpy()
    else:
        raw_data = np.array(df)

    # loop through the windows
    for i in range(num_windows):

        # store X and Y information
        X[i] = raw_data[i: i + train_horizon]
        y[i] = raw_data[i + train_horizon: i + train_horizon + pred_horizon].reshape(1, -1)[0]

    return X, y


def create_multivariate_windows(df, train_horizon, pred_horizon, train_columns=None, pred_columns=None):
    """Given a dataframe or raw data sequence with more than one dimension in time,
     this will create deep learning friendly matrices (X, y) with a given training
     and prediction length. If not columns are provided, the whole matrix will be used
     as the training and prediction set.

    Args:
        df (array_like): raw data potentially with column names if a full dataframe (>1d in time)
        train_horizon (int): number of data points in each training observation of the X matrix
        pred_horizon (int): number of data points in each prediction observation of the y matrix
        train_columns (list, optional): list of columns to use from dataframe for training matrices.
            Defaults to None.
        pred_columns (list, optional): list of columns to use from dataframe for prediction matrices.
            Defaults to None.
    """
    # number of windows
    num_windows = df.shape[0] - train_horizon - pred_horizon + 1

    # create numpy object
    if train_columns is not None:
        raw_X_data = df[train_columns].to_numpy()
        raw_y_data = df[pred_columns].to_numpy()
    else:
        raw_X_data = np.array(df)
        raw_y_data = np.array(df)

    # allocate memory to numpy
    X = np.zeros([num_windows, train_horizon, raw_X_data.shape[1]])
    y = np.zeros([num_windows, pred_horizon, raw_y_data.shape[1]])

    # loop through the windows
    for i in range(num_windows):

        # store X and Y information
        X[i] = raw_X_data[i: i + train_horizon]
        y[i] = raw_y_data[i + train_horizon: i + train_horizon + pred_horizon]

    return X, y



# load libraries
import tensorflow as tf
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from timewarpy import core, datasets

# load and preprocess
df = datasets.load_energy_data()
TSprocessor = core.MultivariateTS(1000, 100, scaler=MinMaxScaler)
X, y = TSprocessor.fit_transform(df, 'Appliances')
X, y = TSprocessor.fit_transform(df, train_columns=['Appliances', 'T1', 'RH_1'], pred_columns=['Appliances', 'T1'])
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33, random_state=42)
print(f'Original dataframe shape: {df.shape}')
print(f'X training vector shape: {X_train.shape}')
print(f'y training vector shape: {y_train.shape}')

# train small tensorflow model
model = tf.keras.models.Sequential()
model.add(tf.keras.layers.LSTM(10, activation='tanh', recurrent_activation='sigmoid', input_shape=X_train[0].shape))
model.add(tf.keras.layers.Dense(200))
model.add(tf.keras.layers.Reshape((100, 2), input_shape=(200,)))
model.compile(optimizer='Adam', loss='mean_squared_error',)
history = model.fit(X_train, y_train, epochs=2, batch_size=100,)

# make predictions
y_pred = model.predict(X_test)
print(f'y prediction vector shape: {y_pred.shape}')



from timewarpy import datasets, core
from sklearn.preprocessing import MinMaxScaler


def test_UnivariateTS_fit():
    df = datasets.load_energy_data()
    TSprocessor = core.UnivariateTS(1680, 240, scaler=MinMaxScaler)
    TSprocessor.fit(df, 'Appliances')
    assert TSprocessor.scaler.data_max_ is not None


def test_UnivariateTS_transform():
    df = datasets.load_energy_data()
    TSprocessor = core.UnivariateTS(1680, 240, scaler=MinMaxScaler)
    TSprocessor.fit(df, 'Appliances')
    X, y = TSprocessor.transform(df, 'Appliances')
    assert X.shape == (17816, 1680, 1)


def test_UnivariateTS_fit_transform():
    df = datasets.load_energy_data()
    TSprocessor = core.UnivariateTS(1680, 240, scaler=MinMaxScaler)
    X, y = TSprocessor.fit_transform(df, 'Appliances')
    assert X.shape == (17816, 1680, 1)


def test_UnivariateTS_inverse_transform():
    df = datasets.load_energy_data()
    TSprocessor = core.UnivariateTS(1680, 240, scaler=MinMaxScaler)
    X, y = TSprocessor.fit_transform(df, 'Appliances')
    y_inv = TSprocessor.inverse_transform(y)
    X_inv = TSprocessor.inverse_transform(X)
    assert y_inv.max() > 1 and X_inv.max() > 1


def test_MultivariateTS_fit():
    df = datasets.load_energy_data()
    TSprocessor = core.MultivariateTS(1680, 240, scaler=MinMaxScaler)
    TSprocessor.fit(df, train_columns=['Appliances', 'T1', 'RH_1'], pred_columns=['Appliances', 'T1'])
    assert TSprocessor.scaler_dict['Appliances'].data_max_ is not None
    assert TSprocessor.scaler_dict['Appliances'].data_max_ != TSprocessor.scaler_dict['T1'].data_max_


def test_MultivariateTS_fit_transform():
    df = datasets.load_energy_data()
    TSprocessor = core.MultivariateTS(1680, 240, scaler=MinMaxScaler)
    X, y = TSprocessor.fit_transform(df, train_columns=['Appliances', 'T1', 'RH_1'], pred_columns=['Appliances', 'T1'])
    assert TSprocessor.scaler_dict['Appliances'].data_max_ is not None
    assert TSprocessor.scaler_dict['Appliances'].data_max_ != TSprocessor.scaler_dict['T1'].data_max_
    assert X.shape == (17816, 1680, 3)



from timewarpy import datasets, preprocess


def test_create_univariate_windows():
    df = datasets.load_energy_data()
    X, y = preprocess.create_univariate_windows(df, 1680, 240, 'Appliances')
    assert X.shape == (17816, 1680, 1)


def test_create_multivariate_windows():
    df = datasets.load_energy_data()
    X, y = preprocess.create_multivariate_windows(
        df, 1680, 240, train_columns=['Appliances', 'T1', 'RH_1'], pred_columns=['Appliances', 'T1']
    )
    assert X.shape == (17816, 1680, 3)
    assert y.shape == (17816, 240, 2)
    X, y = preprocess.create_multivariate_windows(
        df, 1680, 240, train_columns=['Appliances', 'T1', 'RH_1'], pred_columns=['Appliances', ]
    )
    assert X.shape == (17816, 1680, 3)
    assert y.shape == (17816, 240, 1)



from timewarpy import preprocess


# TODO: make base transformer that each inherit from


class UnivariateTS:
    """Core object for processing univariate time series. Use this object
    to set up and save all necessary transformations for pre and post model
    processing of a time series that only contains one dimension outside of time.
    """

    def __init__(self, train_horizon: int, pred_horizon: int, scaler: object = None,
                 roll_increment: int = 0):
        """Initializes the core class. First, this sets the values for how many
        time points should be in the training and forecasting windows. See [here](/#univariate-data)
        for a visual on the training and prediction window lengths. Second, this
        also defines any scaling that needs to occur the variable changing in time.
        Scaling functionality follows the scikit-learn standards for methods needed. See an example
        of a standard scaler [here](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html).

        Args:
            train_horizon (int): number of time steps in each training vector
            pred_horizon (int): number of time steps in each prediction vector
            scaler (object, optional): scaling function to use, usage follows scikit-learn. Defaults to None.
            roll_increment (int, optional): how many time sets to skip while rolling windows. Defaults to 0.
        """
        self.__str__ = 'Univariate Time Series Processing Class'
        self.train_horizon = train_horizon
        self.pred_horizon = pred_horizon
        self.roll_increment = roll_increment
        if scaler is not None:
            self.scaler = scaler
        else:
            self.scaler = None

    def fit(self, df, column):
        """Given a pandas dataframe and column for the univariate
        time series data, this will fit necessary preprocessing to that
        given data column. Currently this is only fitting the scalar to
        the given column in the __init__ function.

        Args:
            df (pandas.DataFrame): univariate time series
            column (str): column to use in the dataframe
        """
        self.column = column
        if self.scaler is not None:
            self.scaler = self.scaler().fit(df[column].to_numpy().reshape(-1, 1))

    def transform(self, df, column):
        """Given a pandas dataframe and column for the univariate
        time series data, tranform the data to a neural network friendly
        set of vectors.

        Args:
            df (pandas.DataFrame): univariate time series
            column (str): column to use in the dataframe

        Returns:
            tuple: X (np.array) training vectors, y (np.array) forecasting/prediction vectors
        """
        time_series = self.scaler.transform(df[[column]])
        X, y = preprocess.create_univariate_windows(
            time_series, self.train_horizon, self.pred_horizon
        )
        return X, y

    def fit_transform(self, df, column):
        """Runs the core.UnivariateTS fit and transform functions in one call.

        Args:
            df (pandas.DataFrame): univariate time series
            column (str): column to use in the dataframe

        Returns:
            tuple: X (np.array) training vectors, y (np.array) forecasting/prediction vectors
        """
        self.fit(df, column)
        X, y = self.transform(df, column)
        return X, y

    def inverse_transform(self, vec):
        """Takes a vector and applies all non-windowing processing inversely
        to get back to the original windowed matrix. This is useful for error
        measurement that is unscaled.

        Args:
            vec (array_like): time series set of vectors (univariate)

        Returns:
            array_like: inverse processes set of vectors
        """

        if len(vec.shape) > 2:
            vec_inv = self.scaler.inverse_transform(vec[:, :, 0])
        else:
            vec_inv = self.scaler.inverse_transform(vec)
        return vec_inv


class MultivariateTS:
    """Core object for processing multivariate time series. Use this object
    to set up and save all necessary transformations for pre and post model
    processing of a time series that only contains one dimension outside of time.
    """

    def __init__(self, train_horizon: int, pred_horizon: int, scaler: object = None,
                 roll_increment: int = 0):
        """Initializes the core class. First, this sets the values for how many
        time points should be in the training and forecasting windows.

        TODO: Need example

        Second, this also defines any scaling that needs to occur the variable changing in time.
        Scaling functionality follows the scikit-learn standards for methods needed. See an example
        of a standard scaler [here](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html).

        Args:
            train_horizon (int): number of time steps in each training vector
            pred_horizon (int): number of time steps in each prediction vector
            scaler (object, optional): scaling function to use, usage follows scikit-learn. Defaults to None.
            roll_increment (int, optional): how many time sets to skip while rolling windows. Defaults to 0.
        """
        self.__str__ = 'Multivariate Time Series Processing Class'
        self.train_horizon = train_horizon
        self.pred_horizon = pred_horizon
        self.roll_increment = roll_increment
        if scaler is not None:
            self.scaler = scaler
        else:
            self.scaler = None

    def fit(self, df, train_columns, pred_columns):
        """Given a pandas dataframe and columns for training and prediction columns in a
        multivariate time series data, this will fit necessary preprocessing to the
        given data columns. Currently this is only fitting the scalar to
        the given column in the __init__ function.

        Args:
            df (pandas.DataFrame): multivariate time series
            train_columns (list): columns for training features to use in the dataframe
            train_columns (list): columns for prediction variables

        """
        self.train_columns = train_columns
        self.pred_columns = pred_columns
        all_cols = set(self.train_columns + self.pred_columns)
        self.scaler_dict = dict(zip(all_cols, [self.scaler() for i in all_cols]))
        if self.scaler is not None:
            for column in all_cols:
                self.scaler_dict[column].fit(df[column].to_numpy().reshape(-1, 1))
        return None

    def transform(self, df, train_columns, pred_columns):
        """Given a pandas dataframe and columns for the multivariate
        time series data, tranform the data to a neural network friendly
        set of vectors.

        Args:
            df (pandas.DataFrame): multivariate time series
            train_columns (list): columns for training features to use in the dataframe
            train_columns (list): columns for prediction variables

        Returns:
            tuple: X (np.array) training vectors, y (np.array) forecasting/prediction vectors
        """
        all_cols = set(self.train_columns + self.pred_columns)
        for column in all_cols:
            df[column] = self.scaler_dict[column].transform(df[column].to_numpy().reshape(-1,1)).T[0]
        X, y = preprocess.create_multivariate_windows(
            df,
            train_horizon=self.train_horizon,
            pred_horizon=self.pred_horizon,
            train_columns=train_columns,
            pred_columns=pred_columns,
        )
        return X, y

    def fit_transform(self, df, train_columns, pred_columns):
        """Runs the core.MultivariateTS fit and transform functions in one call.

        Args:
            df (pandas.DataFrame): multivariate time series
            train_columns (list): columns for training features to use in the dataframe
            train_columns (list): columns for prediction variables

        Returns:
            tuple: X (np.array) training vectors, y (np.array) forecasting/prediction vectors
        """
        self.fit(df, train_columns, pred_columns)
        X, y = self.transform(df, train_columns, pred_columns)
        return X, y

    # TODO: inverse transformation
    # def inverse_transform(self, vec):
    #     return None
